#include "../headers/lex.hpp"
using namespace std;
using enum tok::type;

tok::type getIdentifierCharType(const char& Char) {
    if (isalpha(Char)) return IDENTIFIER;
    else if (isdigit(Char)) return NUMBER;
    else if (Char == '"') return STRING;
    else return UNEXPECTED;
}

vector<tok> lex(const string& in)
{
    vector<tok> resVal;
    for (unsigned int i = 0; i < in.size(); ++i) {
        const char& current = in[i];

        switch (current) {
            case TAG: case DEFINE: case LPAR: case RPAR:
            case LBRACE: case RBRACE: case LBRACKET: case RBRACKET:
            case PLUS: case HYPHEN: case LCOMP: case RCOMP:
            case '$': case '\'': resVal.emplace_back(static_cast<tok::type>(current), string(1, current));
            [[likely]] case ' ': case '\t': case '\r':
            [[likely]] case '\n': break;
            default: {
                tok::type type = getIdentifierCharType(current);
                bool isTypeString = (type == STRING);
                switch (type) {
                    case UNEXPECTED: break;
                    case STRING: ++i;
                    case IDENTIFIER: case NUMBER: {
                        string formedString;
                        for (;i < in.size(); ++i) {
                            tok::type currentCharType = getIdentifierCharType(in[i]);
                            bool isString = currentCharType == STRING;
                            if (i == in.size() - 1 && not isString) throw tok::LexerException("A never ending string was found");
                            if ((currentCharType == type || isTypeString) && !isString) {
                                formedString += string(1, in[i]);
                            } else {
                                if (not isTypeString) --i;
                                resVal.emplace_back(type, formedString);
                                break;
                            }
                        }
                        break;
                    }
                    default: resVal.emplace_back(type, string(1, current));
                }
                break;
            }
        }
    }
    /* Test
    for (const auto& it : resVal) {
        cout << it.toktext << ' ' << it.toktype << '\n';
    }
    */
    return resVal;
}